"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
Object.defineProperty(exports, "mapSavedObjectsType", {
  enumerable: true,
  get: function () {
    return _map_saved_object.mapSavedObjectsType;
  }
});
var _map_saved_object = require("./map_saved_object");
//# sourceMappingURL=data:application/json;charset=utf-8;base64,eyJ2ZXJzaW9uIjozLCJuYW1lcyI6WyJfbWFwX3NhdmVkX29iamVjdCIsInJlcXVpcmUiXSwic291cmNlcyI6WyJpbmRleC50cyJdLCJzb3VyY2VzQ29udGVudCI6WyIvKlxuICogQ29weXJpZ2h0IE9wZW5TZWFyY2ggQ29udHJpYnV0b3JzXG4gKiBTUERYLUxpY2Vuc2UtSWRlbnRpZmllcjogQXBhY2hlLTIuMFxuICovXG5cbmV4cG9ydCB7IG1hcFNhdmVkT2JqZWN0c1R5cGUgfSBmcm9tICcuL21hcF9zYXZlZF9vYmplY3QnO1xuIl0sIm1hcHBpbmdzIjoiOzs7Ozs7Ozs7OztBQUtBLElBQUFBLGlCQUFBLEdBQUFDLE9BQUEiLCJpZ25vcmVMaXN0IjpbXX0=