syntax = "proto3";

message Reserved {
  string x = 1;
  reserved 2, 3;
  reserved 5 to 9;
  reserved "foo", "bar";
  string y = 10;
}

enum ReservedEnum {
  x = 1;
  reserved 2, 3;
  reserved 5 to 9;
  reserved "foo", "bar";
  y = 10;
}
