import type BitVector from "../flat/bitVector";
import Vector from "../vector";
export declare class Int32ConstVector extends Vector<Int32Array | Uint32Array, number> {
    constructor(name: string, value: number, sizeOrNullabilityBuffer: number | BitVector, isSigned: boolean);
    protected getValueFromBuffer(_index: number): number;
}
