message Person {
  enum PhoneType {
    option allow_alias = true;
    MOBILE = 0 [(enum_value_is_deprecated) = true, (some_second_option) = "value1, value2, value3", (some_third_option) = '[value1, value2, value3]'];
    HOME = 1 [(enum_value_is_deprecated) = true];
    WORK = 2;
  }
}
