export declare function fastUnpack32_1(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_2(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_3(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_4(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_5(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_6(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_7(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_8(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_9(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_10(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_11(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_12(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack32_16(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_1(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_2(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_3(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_4(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_5(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_6(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_7(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_8(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_16(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number): void;
export declare function fastUnpack256_Generic(inValues: Uint32Array, inPos: number, out: Uint32Array, outPos: number, bitWidth: number): void;
