import type { GeoJSONVTInternalFeature, GeoJSONVTOptions } from './definitions';
/**
 * converts GeoJSON feature into an intermediate projected JSON vector format with simplification data
 * @param data
 * @param options
 * @returns
 */
export declare function convert(data: GeoJSON.GeoJSON, options: GeoJSONVTOptions): GeoJSONVTInternalFeature[];
//# sourceMappingURL=convert.d.ts.map