//Single-line comment
message Point {
  required int32 x = 1;
  required int32 y = 2;
  optional string label = 3;
}

/*Multi-line comment*/
message Line {/*
*/required Point start = 1;
  required Point end = 2;
  optional string label = 3;
}

/**
 * Doxygen-style comment
 **/
message A {
}//Comment after closing brace of last message
