import type { GeoJSONVTInternalFeature, GeoJSONVTOptions } from './definitions';
export declare function wrap(features: GeoJSONVTInternalFeature[], options: GeoJSONVTOptions): GeoJSONVTInternalFeature[];
//# sourceMappingURL=wrap.d.ts.map