export var ScalarType;
(function (ScalarType) {
    ScalarType[ScalarType["BOOLEAN"] = 0] = "BOOLEAN";
    ScalarType[ScalarType["INT_8"] = 1] = "INT_8";
    ScalarType[ScalarType["UINT_8"] = 2] = "UINT_8";
    ScalarType[ScalarType["INT_32"] = 3] = "INT_32";
    ScalarType[ScalarType["UINT_32"] = 4] = "UINT_32";
    ScalarType[ScalarType["INT_64"] = 5] = "INT_64";
    ScalarType[ScalarType["UINT_64"] = 6] = "UINT_64";
    ScalarType[ScalarType["FLOAT"] = 7] = "FLOAT";
    ScalarType[ScalarType["DOUBLE"] = 8] = "DOUBLE";
    ScalarType[ScalarType["STRING"] = 9] = "STRING";
})(ScalarType || (ScalarType = {}));
//# sourceMappingURL=scalarType.js.map