/**
 * calculate simplification data using optimized Douglas-Peucker algorithm
 * @param coords - flat array of coordinates
 * @param first - index of the first coordinate in the segment
 * @param last - index of the last coordinate in the segment
 * @param sqTolerance - square tolerance value
 */
export declare function simplify(coords: number[], first: number, last: number, sqTolerance: number): void;
//# sourceMappingURL=simplify.d.ts.map