import type { GeometryVector, MortonSettings } from "../vector/geometry/geometryVector";
export declare const DEFAULT_MORTON_SETTINGS: MortonSettings;
export declare function encode(x: number, y: number): number;
export declare function encodePointGeometryVector(x: number, y: number): GeometryVector;
export declare function encodePointGeometryVectorWithOffset(x: number, y: number): GeometryVector;
export declare function encodePointGeometryVectorWithMortonEncoding(x: number, y: number): GeometryVector;
export declare function encodePointsGeometryVector(points: number[]): GeometryVector;
export declare function encodeMultiPointGeometryVector(points: number[][]): GeometryVector;
export declare function encodeLineStringGeometryVector(lines: [number, number][]): GeometryVector;
export declare function encodeLineStringGeometryVectorWithMortonEncoding(line: [number, number][]): GeometryVector;
export declare function encodePolygonGeometryVector(polygon: [number, number][][]): GeometryVector;
export declare function encodePolygonGeometryVectorWithOffsets(polygon: [number, number][][]): GeometryVector;
export declare function encodePolygonGeometryVectorWithMortonOffsets(polygon: [number, number][][]): GeometryVector;
export declare function encodeMultiLineStringGeometryVector(lines: [number, number][][]): GeometryVector;
export declare function encodeMultiLineStringGeometryVectorWithOffsets(lines: [number, number][][]): GeometryVector;
export declare function encodeMultiLineStringGeometryVectorWithMortonOffsets(lines: [number, number][][]): GeometryVector;
export declare function encodeMultiPolygonGeometryVector(polygons: [number, number][][][]): GeometryVector;
export declare function encodeMultiPolygonGeometryVectorWithOffsets(polygons: [number, number][][][]): GeometryVector;
export declare function encodeMultiPolygonGeometryVectorWithMortonOffsets(polygons: [number, number][][][]): GeometryVector;
